# Copyright 2014-present, Apstra, Inc. All rights reserved.
#
# This source code is licensed under End User License Agreement found in the
# LICENSE file at http://www.apstra.com/community/eula

import socket
import datetime
import time

from aeon.exceptions import ProbeError


class BaseDevice(object):
    DEFAULT_PROBE_TIMEOUT = 10

    def __init__(self, target, connector, **kwargs):
        """
        :param target: hostname or ipaddr of target device
        :param kwargs:
            'user' : login user-name, defaults to "admin"
            'passwd': login password, defaults to "admin
        """
        self.target = target
        self.port = kwargs.get('port')
        self.user = kwargs.get('user', 'admin')
        self.passwd = kwargs.get('passwd', 'admin')
        self.timeout = kwargs.get('timeout', self.DEFAULT_PROBE_TIMEOUT)
        self.facts = {}
        self.api = connector(hostname=target, **kwargs)

        if 'no_probe' not in kwargs:
            self.probe()

        if 'no_gather_facts' not in kwargs:
            self.gather_facts()

    def gather_facts(self):
        """
        Will be overridden by subclass
        :return: None
        """
        pass

    def probe(self):
        interval = 1
        start = datetime.datetime.now()
        end = start + datetime.timedelta(seconds=self.timeout)

        port = self.port or socket.getservbyname(self.api.proto)

        while datetime.datetime.now() < end:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.settimeout(interval)
            try:
                s.connect((self.target, int(port)))
                s.shutdown(socket.SHUT_RDWR)
                s.close()
                elapsed = datetime.datetime.now() - start
                return True, elapsed
            except:  # NOQA
                time.sleep(interval)
                pass
        # Raise ProbeError if unable to reach in time allotted
        raise ProbeError('Unable to reach device within %s seconds' % self.timeout)

    def __repr__(self):
        return 'Device(%r)' % self.target

    def __str__(self):
        return '{vendor} {os} at {target}'.format(vendor=self.facts['vendor'],
                                                  os=self.facts['os'],
                                                  target=self.target)



# Copyright 2014-present, Apstra, Inc. All rights reserved.
#
# This source code is licensed under End User License Agreement found in the
# LICENSE file at http://www.apstra.com/community/eula

import re
from lxml import etree
import requests
from requests.auth import HTTPBasicAuth
from copy import copy

import socket

from aeon import exceptions
from aeon.nxos import exceptions as NxosExc


_RE_CONF = re.compile(r"\n\n?\s*")

__all__ = ['NxosConnector']

_NXOS_RESP_XPATH_BODY = 'outputs/output/body'
_NXOS_RESP_XPATH_HTTPCODE = 'outputs/output/code'
_NXOS_RESP_XPATH_CLIERR = 'outputs/output/clierror'


class NxosRequest(object):
    MESSGE_TYPES = ('cli_show', 'cli_show_ascii', 'cli_conf')
    OUTPUT_FMTS = ('json', 'xml')
    CHUNK_VALUES = (0, 1)

    # !! WARNING !! the following string must not contain a leading newline ...
    # !! WARNING !! that is, it must start immediately following the triple-quote

    MESSAGE = """<?xml version="1.0"?>
<ins_api>
<type>{msg_type}</type>
<version>{api_ver}</version>
<chunk>{chunk}</chunk>
<sid>{session_id}</sid>
<input>{command}</input>
<output_format>{resp_fmt}</output_format>
</ins_api>
"""

    DEFAULTS = dict(
        msg_type='cli_show',
        api_ver='0.1',
        chunk=0,
        session_id=1,
        resp_fmt='json',
        command=None
    )

    def __init__(self):
        self.__dict__['params'] = copy(self.DEFAULTS)

    def send(self, api, timeout=None):
        _timeout = timeout if timeout is not None else api.DEFAULT_TIMEOUT
        try:
            resp = requests.post(
                api.api_url, headers=api.api_headers,
                timeout=_timeout,
                auth=api.api_auth, data=str(self))

        except requests.exceptions.ReadTimeout as exc:
            cmd_exc = exceptions.TimeoutError(exc)
            cmd_exc.timeout = timeout
            raise cmd_exc

        except Exception as exc:
            raise NxosExc.RequestError(exc)

        if 401 == resp.status_code:
            cmd_exc = exceptions.UnauthorizedError()
            cmd_exc.message = 'not authorized'
            raise cmd_exc

        if 200 != resp.status_code:
            cmd_exc = NxosExc.CommandError()
            cmd_exc.errorcode = resp.status_code
            cmd_exc.message = "command failed, http_code={0} http_reason={1}".format(
                resp.status_code, resp.reason)
            raise cmd_exc

        return resp

    def __setattr__(self, key, value):
        if key in self.__dict__['params'] and value is not None:
            self.__dict__['params'][key] = value

    def __getattr__(self, item):
        return self.__dict__.get(item) or self.__dict__.get('params').get(item)

    def __str__(self):
        return self.MESSAGE.format(**self.params)


class NxosOperRequest(NxosRequest):
    def __init__(self, command=None):
        super(self.__class__, self).__init__()
        self.command = command


class NxosConfigRequest(NxosRequest):
    def __init__(self):
        super(self.__class__, self).__init__()
        self.msg_type = 'cli_conf'
        self.resp_fmt = 'xml'


class NxosConnector(object):
    DEFAULT_PROTOCOL = 'http'
    DEFAULT_TIMEOUT = 60

    def __init__(self, hostname, **kwargs):
        self.hostname = hostname

        self.proto = kwargs.get('proto') or self.DEFAULT_PROTOCOL
        self.port = kwargs.get('port') or socket.getservbyname(self.proto)

        self.api_url = "{proto}://{target}:{port}/ins".format(
            proto=self.proto, port=self.port, target=self.hostname)

        self.api_headers = {
            'cookie': 'no-cookie',
            'content-type': 'text/xml'
        }

        self.user = kwargs.get('user')
        self.passwd = kwargs.get('passwd')

    # ---------------------------------------------------------------
    #                             PROPERTIES
    # ---------------------------------------------------------------

    @property
    def passwd(self):
        """ passwd is write-only """
        raise RuntimeError("passwd property is write-only")

    @passwd.setter
    def passwd(self, value):
        """ changes the auth user-password """
        self._passwd = value

    @property
    def api_auth(self):
        return HTTPBasicAuth(self.user, self._passwd)

    def exec_config(self, contents, timeout=None):
        rqst = NxosConfigRequest()
        rqst.command = _RE_CONF.sub(' ; ', contents)

        resp = rqst.send(self, timeout)

        # now we need to check the contents of the NX-API response ...
        # we will just check for one instance right now ... may try to gather
        # all of them &| include the XML body as part of the exception at some
        # later point in time.

        as_xml = etree.XML(resp.text)
        cli_error = as_xml.find(_NXOS_RESP_XPATH_CLIERR)
        if cli_error is not None:
            raise NxosExc.CommandError(cli_error.text)

        # config OK, yea!
        return True

    def exec_opcmd(self, command, raw_resp=False, timeout=None, **kwargs):

        rqst = NxosOperRequest(command=command)
        rqst.msg_type = kwargs.get('msg_type')
        rqst.resp_fmt = kwargs.get('resp_fmt')

        resp = rqst.send(self, timeout)

        def rsp_is_xml():
            as_xml = etree.XML(resp.text)
            if raw_resp is True:
                return as_xml

            return as_xml.find(_NXOS_RESP_XPATH_BODY)

        def rsp_is_json():
            as_json = resp.json()
            outputs = as_json['ins_api']['outputs']['output']

            if 'clierror' in outputs:
                raise NxosExc.CommandError(outputs['clierror'])

            return as_json if raw_resp is True else outputs['body']

        proc_rsp = rsp_is_json if 'json' == rqst.resp_fmt else rsp_is_xml
        return proc_rsp()



# Copyright 2014-present, Apstra, Inc. All rights reserved.
#
# This source code is licensed under End User License Agreement found in the
# LICENSE file at http://www.apstra.com/community/eula


class NxosException(Exception):
    pass


class ProbeError(NxosException):
    pass


class RequestError(NxosException):
    """ retains any exception at time of making the command request """
    pass


class NoRespError(NxosException):
    """ indicates that the response does not have a results code """
    pass


class CommandError(NxosException):
    """ indicates the response code != 200 """
    pass



# Copyright 2014-present, Apstra, Inc. All rights reserved.
#
# This source code is licensed under End User License Agreement found in the
# LICENSE file at http://www.apstra.com/community/eula


from aeon.eos.connector import Connector
from aeon.base.device import BaseDevice


__all__ = ['Device']


class Device(BaseDevice):
    OS_NAME = 'eos'

    def __init__(self, target, **kwargs):
        """
        :param target: hostname or ipaddr of target device
        :param kwargs:
            'user' : login user-name, defaults to "admin"
            'passwd': login password, defaults to "admin
        """
        BaseDevice.__init__(self, target, Connector, **kwargs)

    def gather_facts(self):

        facts = self.facts
        got_ver = self.api.execute('show version')

        facts['vendor'] = 'arista'
        facts['os_name'] = self.OS_NAME
        facts['os_version'] = got_ver['version']
        facts['hw_model'] = got_ver['modelName']
        facts['hw_version'] = got_ver['hardwareRevision']
        facts['hw_part_number'] = None
        facts['hw_part_version'] = None
        facts['chassis_id'] = None
        facts['mac_address'] = got_ver['systemMacAddress']

        facts['virtual'] = bool('vEOS' == facts['hw_model'])

        if facts['virtual']:
            got_ma1 = self.api.execute('show interfaces ma1')
            macaddr = got_ma1['interfaces']['Management1']['physicalAddress']
            facts['serial_number'] = macaddr.replace(':', '').upper()
        else:
            facts['serial_number'] = got_ver['serialNumber']

        try:
            got_host = self.api.execute('show hostname')
            facts['fqdn'] = got_host.get('fqdn')
            facts['hostname'] = got_host.get('hostname')
        except:  # NOQA
            facts['fqdn'] = 'localhost'
            facts['hostname'] = 'localhost'



# Copyright 2014-present, Apstra, Inc. All rights reserved.
#
# This source code is licensed under End User License Agreement found in the
# LICENSE file at http://www.apstra.com/community/eula


class TargetError(Exception):
    pass


class ProbeError(Exception):
    pass


class UnauthorizedError(Exception):
    pass


class TimeoutError(Exception):
    pass


class ConfigError(Exception):
    def __init__(self, exc, contents):
        super(ConfigError, self).__init__(exc, contents)
        self.exc = exc
        self.contents = contents


class CommandError(Exception):
    def __init__(self, exc, commands):
        super(CommandError, self).__init__(exc, commands)
        self.exc = exc
        self.commands = commands

    def __repr__(self):
        return self.exc.message


class LoginNotReadyError(Exception):
    def __init__(self, exc, message):
        super(LoginNotReadyError, self).__init__(exc, message)
        self.exc = exc
        self.message = message

    def __repr__(self):
        return self.exc.message



# Copyright 2014-present, Apstra, Inc. All rights reserved.
#
# This source code is licensed under End User License Agreement found in the
# LICENSE file at http://www.apstra.com/community/eula

import re

from aeon.cumulus.connector import Connector
from aeon.base.device import BaseDevice


__all__ = ['Device']


class Device(BaseDevice):
    OS_NAME = 'cumulus'

    def __init__(self, target, **kwargs):
        """
        :param target: hostname or ipaddr of target device
        :param kwargs:
            'user' : login user-name, defaults to "admin"
            'passwd': login password, defaults to "admin
        """
        BaseDevice.__init__(self, target, Connector, **kwargs)

    def _serial_from_link(self, link_name):
        good, got = self.api.execute(['ip link show dev %s' % link_name])
        data = got[0]['stdout']
        macaddr = data.partition('link/ether ')[-1].split()[0]
        return macaddr.upper()

    def gather_facts(self):

        facts = self.facts
        facts['os_name'] = self.OS_NAME

        good, got = self.api.execute([
            'hostname',
            'cat /etc/lsb-release | grep RELEASE | cut -d= -f2',
            'test -e /usr/cumulus/bin/decode-syseeprom'
        ])

        facts['fqdn'] = got[0]['stdout'].strip()
        facts['hostname'] = facts['fqdn']
        facts['os_version'] = got[1]['stdout'].strip()
        virt2 = bool(0 != got[2]['exit_code'])

        if virt2 is True:
            # this is a Cumulus VX 2.x device
            facts['virtual'] = True
            facts['vendor'] = 'CUMULUS-NETWORKS'
            facts['serial_number'] = self._serial_from_link("eth0")
            facts['mac_address'] = self._serial_from_link("eth0")
            facts['hw_model'] = 'CUMULUS-VX'
            facts['hw_part_number'] = None
            facts['hw_version'] = None
            facts['service_tag'] = None
        else:
            good, got = self.api.execute([
                'sudo decode-syseeprom',
            ])

            syseeprom = got[0]['stdout']
            scanner = re.compile(r'(.+) 0x[A-F0-9]{2}\s+\d+\s+(.+)')
            decoded = {
                tag.strip(): value
                for tag, value in scanner.findall(syseeprom)}

            facts['mac_address'] = decoded.get('Base MAC Address', 'no-mac-addr')
            facts['vendor'] = decoded.get('Vendor Name', 'no-vendor-name')
            facts['serial_number'] = decoded.get('Serial Number', 'no-serial-number')
            facts['hw_model'] = decoded.get('Product Name', 'no-product-name')
            facts['virtual'] = bool(facts['hw_model'] == 'VX')
            facts['hw_part_number'] = decoded.get('Part Number', 'no-part-number')
            facts['hw_version'] = decoded.get('Label Revision', 'no-hw-version')
            facts['service_tag'] = decoded.get('Service Tag', 'no-service-tag')



# Copyright 2014-present, Apstra, Inc. All rights reserved.
#
# This source code is licensed under End User License Agreement found in the
# LICENSE file at http://www.apstra.com/community/eula

import pexpect
from pexpect import pxssh

import aeon.nxos.device
import aeon.eos.device
import aeon.cumulus.device
import aeon.ubuntu.device
import aeon.centos.device
from aeon.exceptions import TargetError


def get_device(target=None, user='admin', passwd='admin', nos_only=False):
    """
    Automatically determine device type based on device interrogation.
    :param target: IP address or hostname of target
    :param user: Username to login to target
    :param passwd: Password to login to target
    :param nos_only: Only check for device nos and return as string, do not return device object
    :return: Device object
    """
    dev_table = {
        'nxos': aeon.nxos.device.Device,
        'eos': aeon.eos.device.Device,
        'cumulus': aeon.cumulus.device.Device,
        'ubuntu': aeon.ubuntu.device.Device,
        'centos': aeon.centos.device.Device
    }
    try:
        nos = ''
        session = pxssh.pxssh(timeout=10, options={'StrictHostKeyChecking': "no",
                                       'UserKnownHostsFile': "/dev/null"})
        session.force_password = True
        session.PROMPT = '\r\n.*#|\r\n.*$'
        session.login(target, user, password=passwd, auto_prompt_reset=False, login_timeout=10)

        session.sendline('show version')
        i = session.expect(['Cisco', 'Arista', 'command not found', 'not installed', pexpect.TIMEOUT], timeout=10)
        if i == 0:
            nos = 'nxos'
        if i == 1:
            nos = 'eos'

        if i == 2 or i == 3:
            session.sendline('cat /proc/version')
            i = session.expect(['cumulus', 'Ubuntu', 'Red Hat'], timeout=10)
            if i == 0:
                nos = 'cumulus'
            if i == 1:
                nos = 'ubuntu'
            if i == 2:
                nos = 'centos'
            if i == 3:
                raise TargetError('Unable to determine device type for %s' % target)

    except pxssh.ExceptionPxssh as e:
        raise TargetError("Error logging in: %s" % e)

    finally:
        session.close()

    if nos_only:
        return nos
    else:
        return dev_table[nos](target, user=user, passwd=passwd)



#
#  Copyright 2016-present, Apstra, Inc.  All rights reserved.
#
#  This source code is licensed under Community End User License Agreement
#  found in the LICENSE.txt file in the root directory of this source tree.
#

from setuptools import setup, find_packages
from glob import glob
from os import path


def requirements(filename):
    return filter(None, [
        line.strip()
        for line in open(filename).readlines()])


libdir = 'pylib'
aeondir = 'pylib/aeon'
packages = find_packages(libdir)

setup(
    name="aeon-venos",
    version="0.9.2",
    author="Jeremy Schulman",
    url='https://github.com/Apstra/aeon-venos',
    author_email="jeremy@apstra.com",
    description="Aeon vendor NOS driver library",
    license="Apache 2.0",
    keywords="networking automation vendor-agnostic",
    package_dir={'': libdir},
    packages=packages,
    extras_require={
        "eos": ["pyeapi", "pexpect==4.2.1"],
        "nxos": ["lxml", "requests", "pexpect==4.2.1"],
        "cumulus": ["paramiko<2.0.0", "pexpect==4.2.1"],
        "ubuntu": ["paramiko<2.0.0", "pexpect==4.2.1"],
        "centos": ["paramiko<2.0.0", "pexpect==4.2.1"]
    },
    scripts=glob('bin/*'),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Telecommunications Industry',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Networking',
    ],
)



# Copyright 2014-present, Apstra, Inc. All rights reserved.
#
# This source code is licensed under End User License Agreement found in the
# LICENSE file at http://www.apstra.com/community/eula

__all__ = ['_install_os']


class _install_os(object):
    DESTDIR = 'bootflash'
    VRF_NAME = 'management'

    def __init__(self, device, image=None):
        self.device = device
        self.image = image

    # ##### -------------------------------------------------------------------
    # #####
    # #####                       PROPERTIES
    # #####
    # ##### -------------------------------------------------------------------

    @property
    def md5sum(self):
        """
        Check to see if the file exists on the device
        :return:
        """
        cmd = 'show file {dir}:{bin} md5sum'.format(
            dir=self.DESTDIR, bin=self.image)

        run = self.device.api.exec_opcmd
        try:
            got = run(cmd)
            return got.get('file_content_md5sum').strip()
        except:  # NOQA
            return None

    @property
    def available_space(self):
        cmd = 'df -k /{dir} | grep {dir}'.format(dir=self.DESTDIR)
        run = self.device.api.exec_opcmd
        try:
            got = run(cmd, msg_type='bash')
            return int(got[3])
        except:  # NOQA
            # @@@ TO-DO: need to handle this properly
            raise

    # ##### -------------------------------------------------------------------
    # #####
    # #####                       PUBLIC METHODS
    # #####
    # ##### -------------------------------------------------------------------

    def copy_from(self, location, timeout=10 * 60):
        """
        This method will fetch the image; the fetch will happen from the
        device-side using the 'copy' command.  Note that the NXAPI appears to
        be single-threaded, so the code needs to wait until this operation has
        completed before attempting another API call.  Therefore the :timeout:
        value is set very high (10min)

        :param location: URL to the location of the file.  This URL must be a valid source
        field to the NXOS 'copy' command

        :keyword timeout: Timeout in seconds

        :return:
        """

        cmd = 'copy {location} {dir}: vrf {vrf_name}'.format(
            location=location, dir=self.DESTDIR, vrf_name=self.VRF_NAME)

        run = self.device.api.exec_opcmd
        run(cmd, msg_type='cli_show_ascii', timeout=timeout)

    def run(self, timeout=10 * 60):
        """
        This will invoke the command to install the image, and then
        cause the device to reboot.
        :param timeout: time/seconds to perform the install action
        """

        cmd = 'install all nxos {dir}:{bin}'.format(
            dir=self.DESTDIR, bin=self.image)

        run = self.device.api.exec_opcmd
        run(cmd, msg_type='cli_show_ascii', timeout=timeout)



# Copyright 2014-present, Apstra, Inc. All rights reserved.
#
# This source code is licensed under End User License Agreement found in the
# LICENSE file at http://www.apstra.com/community/eula

from aeon.base.device import BaseDevice
from aeon.cumulus.connector import Connector


__all__ = ['Device']


class Device(BaseDevice):
    OS_NAME = 'centos'

    def __init__(self, target, **kwargs):
        """
        :param target: hostname or ipaddr of target device
        :param kwargs:
            'user' : login user-name, defaults to "admin"
            'passwd': login password, defaults to "admin
        """
        BaseDevice.__init__(self, target, Connector, **kwargs)

    def get_mac_address(self, link_name):
        good, got = self.api.execute(['/sbin/ip link show dev %s' % link_name])
        data = got[0]['stdout']
        macaddr = data.partition('link/ether ')[-1].split()[0]
        return macaddr.upper()

    def gather_facts(self):

        facts = self.facts
        facts['os_name'] = self.OS_NAME
        good, got = self.api.execute([
            'hostname',
            'cat /etc/centos-release | cut -d" " -f3'
        ])

        facts['fqdn'] = got[0]['stdout'].strip()
        facts['hostname'] = facts['fqdn']
        facts['os_version'] = got[1]['stdout'].strip()
        facts['virtual'] = None
        facts['vendor'] = 'CentOS'
        facts['serial_number'] = None
        facts['mac_address'] = self.get_mac_address("eth0")
        facts['hw_model'] = 'Server'
        facts['hw_part_number'] = None
        facts['hw_version'] = None
        facts['service_tag'] = None



# Copyright 2014-present, Apstra, Inc. All rights reserved.
#
# This source code is licensed under End User License Agreement found in the
# LICENSE file at http://www.apstra.com/community/eula

import pyeapi
import socket
from copy import deepcopy

from aeon.exceptions import ConfigError, CommandError


__all__ = ['Connector']


class Connector(object):
    DEFAULT_PROTOCOL = 'http'
    VRF_MGMT = 'management'
    DEFAULT_VRF = 'default'

    def __init__(self, hostname, **kwargs):
        self.vrf = self.VRF_MGMT

        self.hostname = hostname
        self.proto = kwargs.get('proto') or self.DEFAULT_PROTOCOL
        self.port = kwargs.get('port')

        # explicit check for None, since setting port to 0 is
        # valid for on-box EOS.

        if self.port is None:
            self.port = socket.getservbyname(self.proto)

        self.user = kwargs.get('user')
        self.passwd = kwargs.get('passwd')

        if self.proto == 'socket':
            self.eapi = pyeapi.client.make_connection(
                'socket', username=self.user, password=self.passwd,
                **kwargs.get('socket_opts'))
        else:
            self.eapi = pyeapi.connect(
                transport=self.proto, host=self.hostname,
                username=self.user, password=self.passwd)

    def execute(self, commands, encoding='json'):

        # Make a copy of commands so that commands object isn't mutated outside of this function
        commands = deepcopy(commands)

        # Convert to list if not already a list
        commands = commands if isinstance(commands, list) else [commands]

        # Add 'enable' if not the first enty in the commands list
        if commands[0] != 'enable':
            commands.insert(0, 'enable')
        try:
            got = self.eapi.execute(commands=commands, encoding=encoding)
        except Exception as exc:
            raise CommandError(exc=exc, commands=commands)

        # get the results.  if ther was only one command return the actual
        # results item (vs. making the caller do [0]).

        results = got['result']
        results.pop(0)
        return results if len(results) > 1 else results.pop()

    def configure(self, contents):
        if not isinstance(contents, list):
            raise RuntimeError('contents must be a list, for now')

        contents = ['enable', 'configure'] + filter(bool, contents)   # filters out empty lines

        try:
            self.eapi.execute(contents)
        except Exception as exc:
            raise ConfigError(exc=exc, contents=contents)
